﻿using System.ComponentModel;
using System.Diagnostics.CodeAnalysis;
using System.Net.Http.Json;
using System.Text;
using Microsoft.Extensions.Configuration;
using RsaWebapp.Shared.Entities;
using Microsoft.Extensions.Options;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;

namespace RsaWebapp.Shared.Services;


public class FpgaService : IFpgaService
{
    private const string API_PAGE = "/api/";
    private const string APP_SETTINGS_PATH = "../RsaWebapp/appsettings.json";

    private readonly HttpClient _httpClient;
    private readonly IOptions<FpgaConfigurationService> _configurationService;

    public FpgaService(HttpClient httpClient, IOptions<FpgaConfigurationService> configurationService)
    {
        _httpClient = httpClient;
        _configurationService = configurationService;
    }

    //TEMP SOLUTION
    public void ChangeApiValueAtRuntime(FpgaConfigurationService config)
    {
        _configurationService.Value.ApiUrl = config.ApiUrl;
        _configurationService.Value.ApiPort = config.ApiPort;

        var json = File.ReadAllText(APP_SETTINGS_PATH);
        var jsonObj = JObject.Parse(json);
        var fpgaConfig = jsonObj["FpgaConfigurationService"];
        if (fpgaConfig != null)
        {
            fpgaConfig["ApiUrl"] = config.ApiUrl;
            fpgaConfig["ApiPort"] = config.ApiPort;
        }

        File.WriteAllText(APP_SETTINGS_PATH, jsonObj.ToString());

    }

    public async Task<bool> GetConnectionStatus()
    {
        Console.WriteLine($"{_configurationService.Value.ApiUrl}:{_configurationService.Value.ApiPort}");
        try
        {
            Console.WriteLine();
            var response = await _httpClient.GetFromJsonAsync<FpgaStatusResponse>(
                $"http://{_configurationService.Value.ApiUrl}:{_configurationService.Value.ApiPort}{API_PAGE}status");
            if (response == null)
                return false;
            return response.status == "ok" ? true : false;
        }
        catch (Exception ex)
        {
            return false;
        }
    }

    public async Task<EncryptedValue> Encrypt(string message, RsaKeys key)
    {
        var body = new FpgaEncryptBody()
        {
            E = BitConverter.ToInt32(Convert.FromBase64String(key.Exponent)),
            N = BitConverter.ToInt32(Convert.FromBase64String(key.Modulus)),
            Message = message,
        };
        
        var response = await _httpClient.PostAsJsonAsync(
            $"http://{_configurationService.Value.ApiUrl}:{_configurationService.Value.ApiPort}{API_PAGE}encrypt",
            body);

        if (response.IsSuccessStatusCode)
        {
            var responseBody = await response.Content.ReadFromJsonAsync<EncryptionResponse>();
            Console.WriteLine(responseBody.encrypted_msg);
            string[] parts = responseBody.encrypted_msg.Split(' ');
            parts = parts.Skip(1).ToArray();
            
       
            int totalBytesNeeded = 0;
            foreach (string part in parts)
            {
                int decimalValue = Convert.ToInt32(part, 16);
                if (decimalValue >= 256)
                {
                    totalBytesNeeded += 2; 
                }
                else
                {
                    totalBytesNeeded += 1; 
                }
            }

            byte[] byteArray = new byte[totalBytesNeeded];
            int byteIndex = 0;


            foreach (string part in parts)
            {
                int decimalValue = Convert.ToInt32(part, 16);

                if (decimalValue >= 256)
                {
                    byteArray[byteIndex++] = (byte)(decimalValue / 256);
                    byteArray[byteIndex++] = (byte)(decimalValue % 256); 
                }
                else
                {
                    byteArray[byteIndex++] = (byte)decimalValue; 
                }
            }

            return new EncryptedValue
            {
                bytes = byteArray,
                stringVal = responseBody.encrypted_msg,
            };
        }

        return new EncryptedValue();


    }

    public async Task<byte[]> Decrypt(string message, RsaKeys key)
    {
        var body = new FpgaDecryptBody()
        {
            D = BitConverter.ToInt32(Convert.FromBase64String(key.D)),
            N = BitConverter.ToInt32(Convert.FromBase64String(key.Modulus)),
            Message = message
        };
        
        var response = await _httpClient.PostAsJsonAsync(
            $"http://{_configurationService.Value.ApiUrl}:{_configurationService.Value.ApiPort}{API_PAGE}decrypt",
            body);

        if (response.IsSuccessStatusCode)
        {
            try
            {
                var responseBody = await response.Content.ReadFromJsonAsync<DecryptionResponse>();
                Console.WriteLine(responseBody.decrypted_msg);
                byte[] byteList = new byte[responseBody.decrypted_msg.Length / 2];

                for (int i = 0, j = 0; i < responseBody.decrypted_msg.Length; i += 2, j++)
                {
                    string hexPair = responseBody.decrypted_msg.Substring(i, 2);
                    byte hexValue = byte.Parse(hexPair, System.Globalization.NumberStyles.HexNumber);
                    byteList[j] = hexValue;
                }

                return byteList;
            }
            catch
            {
                return null;
            }
        }

        return new byte[0];
    }

    public async Task<RsaKeys> GenerateKey(RsaDetails details)
    {
        try
        {
            var det = new FpgaRsaKeyData
            {
                p = details.PrimeP,
                q = details.PrimeQ,
                e = details.E,
            };
            var response = await _httpClient.PostAsJsonAsync(
                $"http://{_configurationService.Value.ApiUrl}:{_configurationService.Value.ApiPort}{API_PAGE}key",
                det);
            if (response.IsSuccessStatusCode)
            {
                var responseBody = await response.Content.ReadFromJsonAsync<Response>();
                Console.WriteLine(responseBody.nValue);
                return new RsaKeys()
                {
                    P = Convert.ToBase64String(BitConverter.GetBytes(det.p)),
                    Q = Convert.ToBase64String(BitConverter.GetBytes(det.q)),
                    Exponent = Convert.ToBase64String(BitConverter.GetBytes(det.e)),
                    Modulus = Convert.ToBase64String(StringToByteArrayLE(responseBody.nValue)),
                    D = Convert.ToBase64String(StringToByteArrayLE(responseBody.dValue)),
                    Type = GeneratedKeyType.FPGA,
                };
            }
        }
        catch
        {
            Console.WriteLine("nie mozna wygenerowac klucza na fpga");
        }

        return null!;
    }

    public static byte[] StringToByteArrayLE(string hex)
    {
        int numberChars = hex.Length;
        byte[] bytes = new byte[numberChars / 2];
        for (int i = 0; i < numberChars; i += 2)
        {
            // Calculate index based on big-endian order
            int index = (numberChars / 2 - 1) - i / 2;
            bytes[index] = Convert.ToByte(hex.Substring(i, 2), 16);
        }

        return bytes;
    }
}






public class FpgaStatusResponse
{
    public string status { get; set; }
}

public class FpgaRsaKeyData
{
    public int p {  get; set; }
    public int q { get; set; }
    public int e { get; set; }
}

public class Response
{
    public string nValue { get; set; }
    public string dValue { get; set; }
}

public class EncryptionResponse
{
    public string encrypted_msg { get; set;}
}

public class DecryptionResponse
{
    public string decrypted_msg { get; set;}
}

public class FpgaEncryptBody
{
    public string Message { get; set; }
    public int E { get; set; }
    public int N { get; set; }
}

public class FpgaDecryptBody
{
    public string Message { get; set; }
    public int D { get; set; }
    public int N { get; set; }
}

public class EncryptedValue
{
    public byte[] bytes;
    public string stringVal;
}