﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Security.Cryptography;
using RsaWebapp.Shared.Entities;
using System.Xml.Linq;
using System.Xml;
using Microsoft.EntityFrameworkCore.Storage.ValueConversion;
using System.Security.AccessControl;
namespace RsaWebapp.Shared.Services
{
    public class RsaService : IRsaService
    {
        private readonly IFpgaService _fpgaService;
        public RsaService(IFpgaService fpgaService)
        {
            _fpgaService = fpgaService;
        }
        public static int CalculateModularMultiplicativeInverse(int e, int phi)
        {

            int d = 0;
            for (int i = 0; i < phi; i++)
            {
                if ((i * e) % phi == 1)
                {
                    d = i;
                    break;
                }
            }
            return d;
        }

        public static int CalculateModulus(int p, int q)
        {
            return p * q;
        }

        public static int CalculatePhi(int p, int q)
        {
            return (p - 1) * (q - 1);
        }

        public static async Task<List<int>> GetListOfCoprimesAsync(int phi)
        {
            return await Task.Run(() =>
            {
                List<int> coprimes = new List<int>();
                for (int i = 2; i < phi; i++)
                {
                    if (Gcd(i, phi) == 1)
                    {
                        coprimes.Add(i);
                    }
                }
                return coprimes;
            });
        }

        private static int Gcd(int a, int b)
        {
            while (b != 0)
            {
                int temp = b;
                b = a % b;
                a = temp;
            }
            return a;
        }

        public string GenerateKeys(RsaDetails rsaDetails)
        {
            
            RSAParameters rsaParameters= new RSAParameters();
            //rsaParameters.D = BitConverter.GetBytes(rsaDetails.D);
            //rsaParameters.P = BitConverter.GetBytes(rsaDetails.PrimeP);
            //rsaParameters.Q = BitConverter.GetBytes(rsaDetails.PrimeQ);
            rsaParameters.Exponent = BitConverter.GetBytes(rsaDetails.E);

            rsaParameters.Modulus = BitConverter.GetBytes(rsaDetails.N);
            using (RSACryptoServiceProvider RSA = new RSACryptoServiceProvider())
            {
                RSA.ImportParameters(rsaParameters);
                return RSA.ToString();
            }
        }

        public RandomKeyContainer GenerateRandomKeyOnPc()
        {
            RSACryptoServiceProvider RSA = new RSACryptoServiceProvider(512);

            string publicKeyXml = RSA.ToXmlString(false);
            string privateKeyXml = RSA.ToXmlString(true);

            XmlDocument rsaDoc = new XmlDocument();
            rsaDoc.LoadXml(privateKeyXml);

            IEnumerable<XElement> keyParams = new List<XElement>()
            {
            new XElement("Modulus",rsaDoc.SelectSingleNode("//Modulus").InnerText),
            new XElement("Exponent",rsaDoc.SelectSingleNode("//Exponent").InnerText),
            new XElement("P",rsaDoc.SelectSingleNode("//P").InnerText),
            new XElement("Q",rsaDoc.SelectSingleNode("//Q").InnerText),
            new XElement("D",rsaDoc.SelectSingleNode("//D").InnerText),
            };
            return new RandomKeyContainer { PublicKeyXml = publicKeyXml,
                                            PrivateKeyXml = privateKeyXml,
                                            KeyParamsValues = keyParams,
                                            Type =  GeneratedKeyType.PC,
                                            };
        }

        public async Task<RandomKeyContainer> GenerateRandomKeyOnFpga(RsaDetails details)
        {
            var res = await _fpgaService.GenerateKey(details);
            if (res == null)
                return null;
            IEnumerable<XElement> keyParams = new List<XElement>()
            {
                new XElement("Modulus",res.Modulus),
                new XElement("Exponent",res.Exponent),
                new XElement("P",res.P),
                new XElement("Q",res.Q),
                new XElement("D",res.D),
            };

            return new RandomKeyContainer()
            {
                KeyParamsValues = keyParams,
                Type = GeneratedKeyType.FPGA,
            };
        }

        public byte[] EncryptMessageOnPc(string message, RsaKeys key)
        {
            if (message == null)
            {
                throw new ArgumentNullException("message null");
            }
            
            using(RSACryptoServiceProvider RSA = new RSACryptoServiceProvider(512))
            {
                RSA.FromXmlString(key.PublicKeyXml);
                byte[] encryptedBytes = RSA.Encrypt(Encoding.UTF8.GetBytes(message), false);
                return encryptedBytes;
            }

        }

        public async Task<EncryptedValue> EncryptMessageOnFpga(string message, RsaKeys key)
        {
            if (message == null)
            {
                throw new ArgumentNullException("message null");
            }

            var encryptedValue = await _fpgaService.Encrypt(message, key);
            return encryptedValue;
        }
        
        public async Task<byte[]> DecryptMessageOnFpga(string message, RsaKeys key)
        {
            if (message == null)
            {
                throw new ArgumentNullException("message null");
            }

            var decryptedBytes = await _fpgaService.Decrypt(message, key);
            return decryptedBytes;
        }

        public byte[] DecryptMessage(byte[] message, RsaKeys key)
        {
            if (message == null)
            {
                throw new ArgumentNullException("message null");
            }

            try
            {
                using (RSACryptoServiceProvider RSA = new RSACryptoServiceProvider(512))
                {
                    RSA.FromXmlString(key.PrivateKeyXml);
                    byte[] decryptedBytes = RSA.Decrypt(message, false);
                    return decryptedBytes;
                }
            }
            catch (CryptographicException e)
            {
                Console.WriteLine("--------------\n" + e.Message);
            }
            return new byte[1] ;
        }

        public RsaKeys ConvertKeyContainerToRsaKeys(IKeyContainer container)
        {
            string mod = container.KeyParamsValues.First(v => v.Name == "Modulus").Value;
            string exp = container.KeyParamsValues.First(v => v.Name == "Exponent").Value;
            string p   = container.KeyParamsValues.First(v => v.Name == "P").Value;
            string q   = container.KeyParamsValues.First(v => v.Name == "Q").Value;
            string d = container.KeyParamsValues.First(v => v.Name == "D").Value;
            
            RsaKeys key = new RsaKeys { PublicKeyXml = container.PublicKeyXml,
            							PrivateKeyXml = container.PrivateKeyXml,
            							KeyName = "",
            							Modulus = mod,
            							Exponent = exp,
            							P = p,
            							Q = q,
                                        D = d,
                                        Type = container.Type,
                                        };

            if (container is UserKeyContainer)
            {
                UserKeyContainer userKeyContainer = container as UserKeyContainer;
                key.KeyName = userKeyContainer.KeyName;
            }

            return key;
        }
    }
}
