﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.InteropServices.Marshalling;
using System.Security.Cryptography;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using Microsoft.EntityFrameworkCore;
using RsaWebapp.Shared.Data;
using RsaWebapp.Shared.Entities;

namespace RsaWebapp.Shared.Services
{
	public class UserService : IUserService
	{
		private string KEY_DIR = Environment.GetFolderPath(
						 System.Environment.SpecialFolder.DesktopDirectory) + "\\keys\\";
		private readonly AppDbContext _context;
		private readonly IRsaService _rsaService;
		private readonly string _pepper = "AADF314D";
		private readonly int _iteration = 3;

		public UserService(AppDbContext context, IRsaService rsaService)
		{
			_context = context;
			_rsaService = rsaService;
		}

		public async Task<User> Register(RegisterResource resource)
		{
			User user = _context.Users.
							Where(u => u.Username == resource.Username)
							.FirstOrDefault();

			if (user != default)
				return new User { Username = string.Empty };

			user = new User
			{
				Username = resource.Username,
				PasswordSalt = PasswordHasher.GenerateSalt()
			};
			user.PasswordHash = PasswordHasher.ComputeHash(resource.Password, user.PasswordSalt, _pepper, _iteration);
			await _context.Users.AddAsync(user);
			await _context.SaveChangesAsync();

			return user;
		}

		public async Task<User> Login(LoginResource resource)
		{
			var user = await _context.Users
			.FirstOrDefaultAsync(x => x.Username == resource.Username);

			if (user == null)
			{
				return null!;
			}

			var passwordHash = PasswordHasher.ComputeHash(resource.Password, user.PasswordSalt, _pepper, _iteration);
			if (user.PasswordHash != passwordHash)
				return null!;

			return user;
		}

		public async Task<User> GetUser(int id)
		{
			var user = await _context.Users.
				FirstOrDefaultAsync(x => x.Id == id);

			if (user == null)
				return null!;

			return user;
		}

		public async Task<int> SavePcKey(UserKeyContainer container)
		{
			try
			{
				RsaKeys key = _rsaService.ConvertKeyContainerToRsaKeys(container);
				
				var user = await GetUser(container.UserId);
				key.User = user;
				await _context.Keys.AddAsync(key);
				await _context.SaveChangesAsync();
			}
			catch (Exception ex)
			{
				Console.Write(ex.Message);
				return 0;
			}

			return 1;
		}
        public async Task<List<RsaKeys>> GetUserRsaKeys(int userId)
        {
            List<RsaKeys> keys = await _context.Keys.Where(
				rsaKey => rsaKey.User.Id == userId).
	            Include(k => k.User).
	            ToListAsync();

			return keys;

		}

        public async Task<bool> TryDownloadUserKey(int keyId, bool privateKey = false)
        {
	        try
	        {
		        var key = _context.Keys.Where(
			        k => k.Id == keyId).FirstOrDefault();
		        if (key == null)
			        return false;
		        string privateKeyPemString = string.Empty;
		        string publicKeyPemString = string.Empty;
		        using (RSACryptoServiceProvider RSA = new RSACryptoServiceProvider(512))
		        {
			        RSA.FromXmlString(key.PrivateKeyXml);
			        privateKeyPemString = RSA.ExportRSAPrivateKeyPem();

			        RSA.FromXmlString(key.PublicKeyXml);
			        publicKeyPemString = RSA.ExportRSAPublicKeyPem();
		        }

		        if (!Path.Exists(KEY_DIR))
			        Directory.CreateDirectory(KEY_DIR);
		        using (StreamWriter outputFile = new StreamWriter(KEY_DIR + $"publickey_{keyId}.bin"))
		        {
			        outputFile.Write(publicKeyPemString);
		        }

		        using (StreamWriter outputFile = new StreamWriter(KEY_DIR + $"privatekey_{keyId}.bin"))
		        {
			        outputFile.Write(privateKeyPemString);
		        }

		        return true;
	        }
	        catch (Exception exception)
	        {
		        return false;
	        }
        }


        public async Task DeleteUserKey(int keyId)
		{
			await _context.Keys.
				Where(k => k.Id == keyId).
				ExecuteDeleteAsync();	
		}

        public  List<RsaKeys> GetAllRsaKeys()
        {
			List<RsaKeys> keys = _context.Keys.Include(k => k.User).ToList();

            return keys;
        }
    }
    


}
